# -*- coding: utf-8 -*-
"""
Created on Wed Mar 19 09:04:02 2025

@author: Shagedoorn1
"""
import numpy as np
import sympy as sp
import matplotlib.pyplot as plt
def bode_plot(transfer_function, omega_range=np.logspace(-1, 6, 1000), dark_mode=False):
    """
    Generate and display the Bode plot of the given system.
    
    A Bode plot consists of 2 subplots:
        1 magnitude (in decibels) vs frequency (in radians per second)
        2 phase (in degrees) vs frequency (in radians per second)
    
    Parameters:
        transfer_function (TransferFunction):
            A TransferFunction object, see transfer_function.py for details
        omega_range (numpy array):
            A range of frequencies for which the Bode plot is computed,
            default from 10^-1 to 10^6, with 1000 points.
        dark_mode (bool):
            When True, changes the color of the background and lines
            to dark mode. Default False.
    """
    magnitude, phase = transfer_function.calc_magnitude_and_phase(omega_range)
    
    fig_size = (10, 6)
    if dark_mode:
        plt.style.use("dark_background")
        line_width = 3
        font_size = 18
    else:
        line_width = 1.5
        font_size = 12
    
    plt.figure(figsize = fig_size)
    
    plt.subplot(2, 1, 1)
    
    plt.semilogx(omega_range, magnitude, linewidth=line_width, color='cyan' if dark_mode else 'black')
    
    plt.title("Bode plot", fontsize = font_size)
    
    plt.ylabel("Magnitude (dB)", fontsize = font_size)
    
    plt.axhline(linewidth=1, color='white' if dark_mode else 'black')
    
    plt.grid(True, which='both', linestyle="--", linewidth=0.7, color="gray" if dark_mode else 'black')
    
    plt.subplot(2, 1, 2)
    
    plt.semilogx(omega_range, phase, linewidth = line_width, color='magenta' if dark_mode else 'black')
    
    plt.ylabel("phase (deg)", fontsize = font_size)
    plt.xlabel("Frequency (rad/s)", fontsize = font_size)
    
    plt.axhline(linewidth=1, color='white' if dark_mode else 'black')
    plt.axhline(y=-90, linewidth=2, color='white' if dark_mode else 'black', linestyle="--")
    
    plt.grid(True, which="both", linestyle="--", linewidth=0.7, color='gray' if dark_mode else 'black')
    
    plt.show()
def pz_map(transfer_function, dark_mode=False):
    """
    Generate and display the Pole-Zero map of the given system.
    
    The map shows the poles and zeros of the transfer function in the complex
    plane
        - Poles are marked with an "X"
        - Zeros are marked with an "O"
    
    Parameters:
        transfer_function (TransferFunction):
            The transfer_function object that represents the system,
            see transfer_function.py for details.
        dark_mode (bool):
            When True, changes the color of the background and lines
            to dark mode. Default False.
    
    """
    fig_size = (10, 6)
    
    poles = transfer_function.poles
    zeros = transfer_function.zeros
    
    i = 0
    j = 0
    
    reals = []
    comps = []
    
    if dark_mode:
        plt.style.use('dark_background')
        color = 'magenta'
        h_color = 'gray'
        g_color = 'white'
        font_size = 18
    else:
        color = 'black'
        h_color = 'black'
        g_color = 'black'
        font_size = 12
    
    plt.figure(figsize=fig_size)
    plt.title("Pole-Zero map", fontsize=font_size)
    
    plt.scatter(
        [sp.re(p) for p in poles],
        [sp.im(p) for p in poles],
        marker = 'x',
        s = 500,
        color = color,
        label = 'poles')
    
    while i < len(zeros):
        reals.append(sp.re(zeros[i]))
        comps.append(sp.im(zeros[i]))
        i += 1
    
    plt.scatter(
        [sp.re(z) for z in zeros],
        [sp.im(z) for z in zeros],
        marker = "o",
        s = 500,
        edgecolors = color,
        facecolors = 'none',
        label = 'zeros')
    
    while j < len(poles):
        reals.append(sp.re(poles[j]))
        comps.append(sp.im(poles[j]))
        j += 1
    
    plt.xlim([float(min(reals)) - 1, float(max(reals)) + 1])
    plt.ylim([float(min(comps))-1, float(max(comps)) + 1])
    
    plt.axhline(linewidth=2, color=h_color)
    plt.axvline(linewidth=2, color=h_color)
    
    plt.xlabel(r"$\lambda$ re(s)", fontsize=font_size)
    plt.ylabel(r"$j \omega$ im(s)", fontsize=font_size)
    
    plt.legend(prop={'size': 15})
    plt.grid(True, which='both', linestyle="--", linewidth=0.7, color=g_color)
    
    plt.show()
    
def step_response(transfer_function, dark_mode=False):
    """
    Generate and display the step response of the given system.
    
    The step response represents how a system reacts through time after a unit
    step signal is supplied. A unit step has value 0 when t < 0 and
    value 1 when t > 0
    
    Parameters:
        transfer_function (TransferFunction):
            A TransferFunction object, see transfer_function.py for details
        dark_mode (bool):
            When True, changes the color of the background and lines
            to dark mode. Default False.
    """
    fig_size = (10, 6)
    if dark_mode:
        plt.style.use('dark_background')
        l_color1 = 'magenta'
        l_color2 = 'pink'
        h_color = 'gray'
        g_color = 'white'
        font_size = 18
    else:
        l_color1 = 'black'
        l_color2 = 'red'
        h_color = 'black'
        g_color = 'black'
        font_size = 12
    
    step_in, step_out = transfer_function.step_response()
    
    t = np.linspace(-1, 10, 1000)               #Start before step, end after stabilization
    
    step_in_num = np.array([step_in.subs(transfer_function.var, x).evalf() for x in t])
    step_out_num = np.array([step_out.subs(transfer_function.var, x).evalf() for x in t])
    
    plt.figure(figsize=fig_size)
    plt.title("Step Response", fontsize=font_size)
    
    plt.plot(t, step_in_num, color = l_color1, linewidth = 3, label = '$y_{in}$')
    plt.plot(t, step_out_num, color = l_color2, linewidth = 3, label = '$y_{out}$')
    
    plt.axhline(linewidth = 2, color = h_color)
    plt.axvline(linewidth = 2, color = h_color)
    
    plt.xlabel('$time$ [s]', fontsize = font_size)
    plt.ylabel('$y(t)$', fontsize = font_size)
    
    plt.grid(True, which='both', linestyle="--", linewidth=0.7, color=g_color)
    plt.legend(prop={'size': 15})
    
    plt.show()

def nyquist(transfer_function, dark_mode=False):
    """
    Generate and display the Nyquist diagram of the given system.
    
    The Nyquist diagram is a graphic representation of the systems frequency
    response, plotted in the complex plane. It is used to determine the systems
    stability and performance.
    
    Parameters:
        transfer_function (TransferFunction):
            A TransferFunction object, see transfer_function.py for details
        dark_mode (bool):
            When True, changes the color of the background and lines
            to dark mode. Default False.
    """
    fig_size = (10, 6)
    if dark_mode:
        plt.style.use('dark_background')
        l_color = 'magenta'
        h_color = 'gray'
        g_color = 'white'
        font_size = 18
    else:
        l_color = 'black'
        h_color = 'black'
        g_color = 'black'
        font_size = 12
    
    #Make sure to be in the Fourier domain
    if transfer_function.domain:
        transfer_function.switch_domain()
    #closed = transfer_function.close_loop()
    func = sp.lambdify(transfer_function.w, transfer_function.transfer_func, 'numpy')
    
    omega_pos = np.logspace(-2, 5, 5000)
    omega_neg = -omega_pos[::-1]
    
    func_pos = func(omega_pos)
    func_neg = func(omega_neg)              #Account for negative frequencies
    
    func_vals = np.concatenate((func_neg, func_pos))
    
    func_re = np.real(func_vals)
    func_im = np.imag(func_vals)
    
    plt.figure(figsize=fig_size)
    plt.title("Nyquist diagram", fontsize=font_size)
    
    plt.plot(func_re, func_im, color=l_color, linewidth = 3)
    
    plt.axhline(linewidth=2, color=h_color)
    plt.axvline(linewidth=2, color=h_color)
    
    plt.xlabel("$re(H)$", fontsize=font_size)
    plt.ylabel("$im(H)$", fontsize=font_size)
    
    plt.grid(True, which='both', linestyle="--", linewidth=0.7, color=g_color)
    
    plt.show()
    

if __name__ == "__main__":
    from .transfer_function import TransferFunction
    string1 = "y''+y'+y"
    string2 = "x'+x"
    o_range = np.logspace(-1,3,1000)
    Trans = TransferFunction(string1, string2)
    #bode_plot(Trans, np.logspace(-1,3,1000), True)
    #pz_map(Trans, True)
    #step_response(Trans, True)
    nyquist(Trans, True)


from .Diffeq import DiffEq
from .laplace_handler import LaplaceHandler
from .transfer_function import TransferFunction
from .solver import Solver
from .plotting import bode_plot, pz_map, step_response, nyquist
from .PDiffeq import PDiffEq
__all__ = ["DiffEq", "LaplaceHandler", "TransferFunction", "Solver", "bode_plot", "pz_map", "step_response", "nyquist", "PDiffEq"]


# -*- coding: utf-8 -*-
"""
Created on Thu Apr  3 16:36:01 2025

@author: Shagedoorn1
"""

import sympy as sp
import numpy as np
import matplotlib.pyplot as plt

class Solver:
    """
    A class for numerically solving first order ordinary differential equations.
    
    An ODE is entered as a string in Lagrange notation. The ODE is set to 0 and
    solved for the derivative of y. The result is lambdified and can be
    numerically solved.
    """
    def __init__(self, diff_eq, times, plot=False, dark_mode=False):
        """
        Initialize the solver class.
        
        Arguments:
            diff_eq (DiffEq):
                A DiffEq object, see Diffeq.py for details.
            times (list or tuple):
                A list or tuple with the start and stop time for the numeric
                solution.
            plot (bool):
                When True, a plot is made when the given equation is
                numerically solved, default False.
            dark_mode (bool):
                Only relevant if plot is True. When True, the plot made by the
                numerical solvers is in dark mode.
        Parameters:
            start (float):
                The start time of the numerical solvers.
            stop (float):
                The stop time of the numerical solvers.
            x0 (float):
                The initial condition of the ODE.
                
        """
        self.diff_eq = diff_eq
        self.times = times
        self.start = self.times[0]
        self.stop = self.times[1]
        self.plot = plot
        self.dark_mode = dark_mode
        self.eq = self.diff_eq.isolate()
        self.x0 = self.diff_eq.initial_conds()[0]
        self.time = [self.start]
        self.fig_size = (10,15)
        if self.dark_mode:
            self.line_width = 3
            self.color = 'magenta'
            self.font_size = 18
        else:
            self.line_width = 1.5
            self.color = 'black'
            self.font_size = 12
            
    
    def euler(self, n=200):
        """
        Solves the given ODE with the Euler method.
        
        Arguments:
            n (int):
                The amount of time steps over which the ODE is solved.
                Default 200.
        Returns:
            x (numpy array):
                An array containing the first order solution of the equation.
        """
        n = int(n)
        dt = (self.stop - self.start) / n
        x = np.zeros(n + 1)
        x[0] = self.x0
        t = self.start
        for i in range(1, n+1):
            x[i] = x[i-1] + dt * self.eq(t, x[i-1])
            t += dt
            self.time.append(t)
        if self.dark_mode:
            plt.style.use("dark_background")
        else:
            plt.style.use('default')
        if self.plot:
            plt.figure(figsize = self.fig_size)
            plt.title(f"{self.diff_eq.function} = 0", fontsize = self.font_size)
            plt.plot(
                self.time,
                x,
                linewidth = self.line_width,
                label = "Euler",
                color = self.color,
                )
            plt.grid()
            plt.legend()
            plt.show()
            return x
    
    def RK2(self, n=200):
        """
        Solves the given ODE with the Runge Kutta 2 method.
        
        Arguments:
            n (int):
                The amount of time steps over which the ODE is solved.
                Default 200.
        Returns:
            x (numpy array):
                An array containing the second order solution of the equation.
        """
        dt = (self.stop - self.start) / n
        x = np.zeros(n + 1)
        x[0] = self.x0
        t = self.start
        for i in range(1, n+1):
            k1 = dt * self.eq(t, x[i-1])
            k2 = dt * self.eq(t + 0.5 * dt, x[i-1] + 0.5 * k1)
            x[i] = x[i-1] + k2
            t += dt
            self.time.append(t)
        if self.dark_mode:
            plt.style.use("dark_background")
        else:
            plt.style.use("default")
        if self.plot:
            plt.figure(figsize = self.fig_size)
            plt.title(f"{self.diff_eq.function} = 0", fontsize = self.font_size)
            plt.plot(
                self.time,
                x,
                linewidth = self.line_width,
                label = "Runge Kutta 2",
                color = self.color,
                )
            plt.grid()
            plt.legend()
            plt.show()
            return x
        
    def RK4(self, n=200):
        """
        Solves the given ODE with the Runge Kutta 4 method.
        
        Arguments:
            n (int):
                The amount of time steps over which the ODE is solved.
                Default 200.
        Returns:
            x (numpy array):
                An array containing the fourth order solution of the equation.
        """
        dt = (self.stop - self.start) / n
        x = np.zeros(n + 1)
        x[0] = self.x0
        t = self.start
        for i in range(1, n+1):
            k1 = dt * self.eq(t, x[i-1])
            k2 = dt * self.eq(t + 0.5 * dt, x[i-1] + 0.5 * k1)
            k3 = dt * self.eq(t + 0.5 * dt, x[i-1] + 0.5 * k2)
            k4 = dt * self.eq(t + dt, x[i-1] + k3)
            x[i] = x[i-1] + (1/6) * (k1 + 2 * k2 + 2 * k3 + k4)
            t += dt
            self.time.append(t)
        if self.dark_mode:
            plt.style.use("dark_background")
        else:
            plt.style.use("default")
        if self.plot:
            plt.figure(figsize = self.fig_size)
            plt.title(f"{self.diff_eq.function} = 0", fontsize = self.font_size)
            plt.plot(
                self.time,
                x,
                linewidth = self.line_width,
                label = "Runge Kutta 4",
                color = self.color,
                )
            plt.grid()
            plt.legend()
            plt.show()
            return x
        
if __name__ == "__main__":
    from .Diffeq import DiffEq
    string = "y'+y"
    D = DiffEq(string, 'y')
    Solve = Solver(D, [0, 2], plot = True, dark_mode = True)
    #Solve.euler(10)
    #Solve.RK2(10)
    Solve.RK4(10)


# -*- coding: utf-8 -*-
"""
Created on Tue Apr  8 10:58:43 2025

@author: 23066776
"""

import sympy as sp
import math
sp.init_printing()

class PDiffEq:
    """
    A class for handling partial differential equations (PDEs).
    This class allows users to enter a PDE in "extended" lagrange notation.
    This class assumes a function:
        y(t, x)
    The extended lagrange notation goes as follows:
        ∂y/∂t = f'
        ∂y/∂x = f.
        ∂2y/∂t2 = f''
        ∂2y/∂x2 = f..
        ∂2y/∂t∂x = f'.
    """
    def __init__(self, eq, name="y", var1="t", var2="x"):
        """
        Initializes the PDiffEq class.
        
        Parameters:
            eq (str):
                The equation which is to be converted.
            name (str):
                The name of the function in the equation.
            var1 (str):
                The first variable of the function.
            var2 (str):
                The second variable of the function.
        
        Attributes:
            eq:
                The input equation with whitespaces removed
            terms (list):
                A list containing individual terms of the equation.
                Terms are separated by the '+' operator.
            var1 (str):
                The first independent variable.
            var2 (str):
                The second independent variable
            t (sympy Symbol):
                Symbolic representation of the first independent variable.
            x (sympy Symbol):
                Symbolic representation of the second independent variable.
            name (str):
                The function name.
            y (sympy Function):
                Symbolic representation of the function dependend on 't'.
            tprims (list):
                A list indicating the order of derivatives to the first 
                variable in each term.
                See 'find_derivs()' for details.
            xprims (list):
                A list indicating the order of derivatives to the second 
                variable in each term.
            K (list):
                A list that contains the numerical coefficients of each term.
            E (list):
                A list that contains the numerical exponents of each term.
            function (sympy expression):
                The final converted differential equation as a sympy expression.
        """
        self.eq = eq.replace(" ", "")
        
        self.var1 = var1.lower()
        self.var2 = var2.lower()
        self.name = name.lower()
        
        self.t = sp.Symbol(var1)
        self.x = sp.Symbol(var2)
        self.y = sp.Function(self.name)(self.t, self.x)
        
        self.terms = self.eq.split("+")
        
        self.tprims, self.xprims = self.find_derivs()
        
        self.K, self.E = self.find_factors()
        
        self.equ = self.make_equation()
        self.function = sum(self.equ)
    
    def find_derivs(self):
        """
        A function to find the derivatives to the variables in the given
        equation.
        Returns:
            tprims (list):
                A list containing the derivatives to t for each term.
            xprims (list):
                A list containing the derivatives to x for each term.
        """
        tprims = []
        xprims = []
        
        for term in self.terms:
            tcount = term.count("'")
            xcount = term.count(".")
            
            if (self.name in term):
                if tcount != 0 or xcount != 0:
                    tval = tcount if tcount != 0 else "A"
                    xval = xcount if xcount != 0 else "A"
                else:
                    tval = xval = "A"
                    
            else:
                if tcount != 0 or xcount != 0:
                    raise ValueError
                tval = xval = 0
            tprims.append(tval)
            xprims.append(xval)
        return tprims, xprims
    
    def find_factors(self):
        """
        Find the numerical factors and exponents of each term in the equation.
        
        Returns:
            Ks (list):
                A list with the numerical factor of each term.
            es (list):
                A list with the numerical exponent of each term
        """
        fact = []
        exps = []
        for term in self.terms:
            i = term.split("*")
            sub_l = []
            exp = []
            for char in i:
                if "^" in char:
                    j = char.split("^")
                    for m in j:
                        try:
                            e = int(m)
                            exp.append(e)
                        except ValueError:
                            exp.append(1)
                try:
                    f = int(char)
                    sub_l.append(f)
                except ValueError:
                    if "-" in char:
                        sub_l.append(-1)
                    else:
                        sub_l.append(1)
            fact.append(sub_l)
            exps.append(exp)
            Ks = [math.prod(sub) for sub in fact]
            es = [math.prod(sub) for sub in exps]
        return Ks, es
    
    def make_equation(self):
        """
        Build the PDE from the derivatives, factors and exponents.
        """
        equ = []
        for K, Yt, Yx, T, E in zip(self.K, self.tprims, self.xprims, self.terms, self.E):
            if Yt == "A" and Yx == "A":
                term = K * self.y**E
            elif (Yt != "A" and Yt > 0) and (Yx != "A"and Yx > 0):
                term = K * sp.diff(self.y, self.t, Yt, self.x, Yx)**E
            elif (Yt != "A" and Yt > 0) or (Yx != "A" and Yx > 0):
                term = K * sp.diff(self.y, self.t, Yt if (Yt != "A" and Yt > 0) else 0, self.x, Yx if (Yx != "A" and Yx > 0) else 0)
            else:
                term = sp.sympify(T)
            equ.append(term)
        return equ

if __name__ == "__main__":
    sting = "y.'+y'+y.+3*t+y"
    a = PDiffEq(sting)
    a.function